import argparse
import random
import json
from pathlib import Path
from typing import TextIO, Iterable, Optional

from curricula.shell.plugin import Plugin, PluginDispatcher, PluginException
from curricula.library.serialization import dump
from curricula.log import log

from .run import run, run_batch
from .models import GradingAssignment
from .grader.report import AssignmentReport


def make_report_name(target_path: Path, extension: str) -> str:
    """Generate a report file name."""

    return f"{target_path.parts[-1]}.report.{extension}"


def change_extension(report_path: Path, extension: str) -> str:
    """Return the report name with a different extension."""

    basename = report_path.parts[-1].rsplit(".", maxsplit=1)[0]
    return f"{basename}.{extension}"


def dump_report(report: AssignmentReport, file: TextIO, thin: bool = False):
    """Write a dict of reports to a file."""

    dump(report.dump(thin=thin), file, indent=2)


def amend_report(
        assignment: GradingAssignment,
        existing_report_path: Path,
        new_report: AssignmentReport) -> AssignmentReport:
    """Amend an existing report and return the merged result."""

    with existing_report_path.open() as file:
        existing_report = AssignmentReport.load(json.load(file), assignment)
    for problem_short, problem_report in new_report.problems.items():
        for task_name, result in problem_report.automated.results.items():
            existing_report[problem_short].automated.results[task_name] = result
    return existing_report


def path_from_options(
        options: dict,
        default_file_name: str,
        *,
        batch: bool,
        required: bool = True) -> Optional[Path]:
    """Return an open file and whether to close it after."""

    if options["file"] is not None:
        if batch:
            raise PluginException("Cannot use --file for batch grading, use --directory")
        output_path = Path(options["file"])
        if not output_path.parent.exists():
            raise PluginException(f"Containing directory {output_path.parent} does not exist")
        return output_path
    elif options["directory"] is not None:
        return Path(options["directory"]).joinpath(default_file_name)
    elif required:
        raise PluginException("Output file or directory must be specified!")

    return None


class GradeRunPlugin(Plugin):
    """For running submissions."""

    name = "run"
    help = "run submissions against a grading artifact"

    def setup(self, parser: argparse.ArgumentParser):
        """Command line."""

        parser.add_argument("--grading", "-g", required=True, help="the grading artifact")
        parser.add_argument("--skip", action="store_true", help="skip reports that have already been run")
        parser.add_argument("--report", "-r", action="store_true", help="whether to log test results")
        parser.add_argument("--concise", "-c", action="store_true", help="whether to report concisely")
        parser.add_argument("--progress", "-p", action="store_true", help="show progress in batch")
        parser.add_argument("--sample", type=int, help="randomly sample batch")
        parser.add_argument("--tags", nargs="+", help="only run tasks with the specified tags")
        parser.add_argument("--tasks", nargs="+", help="only run the specified tasks")
        parser.add_argument("--summarize", action="store_true", help="summarize after running batch")
        parser.add_argument("--thin", action="store_true", help="shorten output for space")
        parser.add_argument("--amend", action="store_true", help="amend any existing report at the destination")
        parser.add_argument("--format", action="store_true", help="also format generated reports")
        parser.add_argument("--format-template", type=str, help="the template directory to generate reports from")
        to_group = parser.add_mutually_exclusive_group(required=True)
        to_group.add_argument("-f", "--file", dest="file", help="output file for single report")
        to_group.add_argument("-d", "--directory", dest="directory", help="where to write reports to if batched")
        parser.add_argument("submissions", nargs="+", help="run tests on a single target")

    def main(self, parser: argparse.ArgumentParser, arguments: dict) -> int:
        """Run the grader."""

        grading_path: Path = Path(arguments["grading"]).absolute()
        if not grading_path.is_dir():
            raise PluginException("grading artifact does not exist!")

        if len(arguments["submissions"]) == 1:
            submission_path = Path(arguments["submissions"][0]).absolute()
            return self.run_single(grading_path, submission_path, arguments)
        else:
            return self.run_batch(grading_path, map(Path, arguments.get("submissions")), arguments)

    @classmethod
    def run_single(cls, grading_path: Path, assignment_path: Path, options: dict) -> int:
        """Grade a single file, write to file output."""

        log.debug(f"running single target {assignment_path}")
        assignment = GradingAssignment.read(grading_path)

        report = run(assignment, assignment_path, options=options)
        report_path = path_from_options(options, make_report_name(assignment_path, "json"), batch=False)
        if options["amend"] and report_path.is_file():
            report = amend_report(assignment, report_path, report)

        with report_path.open("w") as file:
            dump_report(report, file, thin=options.get("thin"))

        if options.get("format"):
            cls.format(grading_path, assignment, report_path, options)

        return 0

    @classmethod
    def run_batch(cls, grading_path: Path, target_paths: Iterable[Path], options: dict) -> int:
        """Run a batch of reports."""

        log.debug("running batch submissions")

        assignment = GradingAssignment.read(grading_path)

        # Check which reports have already been run if flagged
        if options["skip"]:
            log.debug("determining reports to skip")
            new_target_paths = []
            skip_count = 0
            for target_path in target_paths:
                report_path = path_from_options(options, make_report_name(target_path, "json"), batch=True)
                if not report_path.parent.exists():
                    log.info(f"making report directory {report_path.parent}")
                    report_path.parent.mkdir(parents=True)
                if not report_path.exists():
                    new_target_paths.append(target_path)
                else:
                    skip_count += 1
            log.info(f"found {skip_count} reports to skip")
            target_paths = new_target_paths

        target_paths = tuple(target_paths)

        # Do random sampling
        sample = options.get("sample")
        if sample is not None:
            target_paths = random.sample(target_paths, sample)

        report_paths = []
        for i, (target_path, report) in enumerate(run_batch(assignment, target_paths, options=options)):
            report_path = path_from_options(options, make_report_name(target_path, "json"), batch=True)
            if options["amend"] and report_path.is_file():
                report = amend_report(assignment, report_path, report)

            with report_path.open("w") as file:
                dump_report(report, file)
            report_paths.append(report_path)

            # Print progress
            if options.get("progress"):
                print(f"{i + 1}/{len(target_paths)} graded")

            if options.get("format"):
                cls.format(grading_path, assignment, report_path, options)

        if options.get("summarize"):
            from curricula_grade.tools.summarize import summarize
            summarize(grading_path, report_paths)

        return 0

    @classmethod
    def format(cls, grading_path: Path, assignment: GradingAssignment, report_path: Path, options: dict):
        from curricula_format import format_report_markdown, create_format_environment

        custom_template_path = Path(options.get("format_template") or grading_path.joinpath("report.md"))
        environment = create_format_environment(custom_template_path=custom_template_path)

        formatted_report_path = report_path.parent.joinpath(change_extension(report_path, "md"))
        formatted_report = format_report_markdown(
            assignment=assignment,
            report_path=report_path,
            environment=environment,
            options=options)
        with formatted_report_path.open("w") as file:
            file.write(formatted_report)


class GradeSummarizePlugin(Plugin):
    """Summarize the results of a set of reports."""

    name = "summarize"
    help = "summarize the results of a set of reports"

    def setup(self, parser: argparse.ArgumentParser):
        """No options."""

        parser.add_argument("--grading", "-g", required=True, help="the grading artifact")
        parser.add_argument("reports", nargs="+", help="the reports to summarize")

    def main(self, parser: argparse.ArgumentParser, arguments: dict):
        """Summarize a batch of reports."""

        grading_path: Path = Path(arguments["grading"]).absolute()
        if not grading_path.is_dir():
            raise PluginException("grading artifact does not exist!")

        from .tools.summarize import summarize
        summarize(grading_path, list(map(Path, arguments["reports"])))


class GradeDiagnosePlugin(Plugin):
    """Get formatted diagnostics on a single report."""

    name = "diagnose"
    help = "get formatted diagnostics on a single report"

    def setup(self, parser: argparse.ArgumentParser):
        """No arguments."""

        parser.add_argument("report", help="report to print diagnostics on")

    def main(self, parser: argparse.ArgumentParser, arguments: dict):
        """Get diagnostics on a report."""

        grading_path: Path = Path(arguments["grading"]).absolute()
        if not grading_path.is_dir():
            raise PluginException("grading artifact does not exist!")

        assignment = GradingAssignment.read(grading_path)
        report_path = Path(arguments["report"])

        from curricula_grade.tools.diagnostics import get_diagnostics
        print(get_diagnostics(assignment, report_path), end="")


class GradePlugin(PluginDispatcher):
    """Implement grade plugin."""

    name = "grade"
    help = "manage assignment grading for submissions"
    plugins = (
        GradeRunPlugin(),
        GradeSummarizePlugin(),
        GradeDiagnosePlugin())

    @classmethod
    def compare(cls, grading_path: Path, options: dict):
        """Generate a comparison of two files."""

        from curricula_grade.tools.compare import compare_output

        report_path = Path(options.get("report"))
        template_path = Path(options.get("template"))
        with path_from_options(options, change_extension(report_path, "compare.html"), batch=False).open("w") as file:
            file.write(compare_output(template_path, report_path))



from typing import Iterator

from . import Task
from ...exception import GraderException
from ...grader.report import ProblemReport

from typing import Dict, List

__all__ = ("topological_sort", "fulfills_dependencies", "flatten_dependencies")


def topological_sort_visit(task: Task, lookup: Dict[str, Task], marks: Dict[Task, int], result: List[Task]):
    """Visit a node."""

    if marks[task] == 2:
        return

    if marks[task] == 1:
        raise GraderException("found cycle in task dependencies")

    marks[task] = 1
    for dependency in task.dependencies.passing | task.dependencies.complete:
        topological_sort_visit(lookup[dependency], lookup, marks, result)
    marks[task] = 2
    result.append(task)


def topological_sort(tasks: List[Task]):
    """Order tasks by dependency."""

    lookup = {}
    marks = {}
    result = []
    for task in tasks:
        marks[task] = 0
        lookup[task.name] = task
    for task in tasks:
        if marks[task] != 2:
            topological_sort_visit(task, lookup, marks, result)
    tasks.clear()
    tasks.extend(result)


def fulfills_dependencies(task: Task, report: ProblemReport):
    """Convenience."""

    return all((
        all(report.automated[dependency].passing for dependency in task.dependencies.passing),
        all(report.automated[dependency].complete for dependency in task.dependencies.complete)))


def flatten_dependencies(task_name: str, task_lookup: Dict[str, Task]) -> Iterator[str]:
    """Return a flattened iterable of dependencies task names."""

    for related_task_name in task_lookup[task_name].dependencies.all():
        yield related_task_name
        yield from flatten_dependencies(related_task_name, task_lookup)



import abc
from typing import Any, Optional, Type

from ..grader.task import Result

__all__ = (
    "Executor",
    "Connector",
    "Evaluator",
    "Runnable",
    "CompositeRunnable",)


class Runnable(metaclass=abc.ABCMeta):
    """Convenience class for building dynamic test objects."""

    @property
    @abc.abstractmethod
    def result_type(self) -> Type[Result]:
        """Class-based tests must indicate a result type."""

    @abc.abstractmethod
    def __call__(self, **kwargs) -> Type[Result]:
        """Must implement runnable invocation."""


class Executor(metaclass=abc.ABCMeta):
    """Runs the content of the test."""

    result_type: Type[Result]
    resources: dict
    details: dict

    @abc.abstractmethod
    def execute(self) -> Any:
        """Does something and produces output."""


class Connector(metaclass=abc.ABCMeta):
    """Converts the result of the executor."""

    result_type: Type[Result]
    resources: dict
    details: dict

    @abc.abstractmethod
    def connect(self, result: Any) -> Any:
        """Transforms output for evaluation."""


class Evaluator(metaclass=abc.ABCMeta):
    """Evaluates the content of a test."""

    result_type: Type[Result]
    resources: dict
    details: dict

    @abc.abstractmethod
    def evaluate(self, result: Any) -> Result:
        """Evaluate and return a result."""


class CompositeRunnable(Runnable, metaclass=abc.ABCMeta):
    """Separate responsibilities."""

    resources: Optional[dict]
    details: Optional[dict]

    @abc.abstractmethod
    def execute(self) -> Any:
        """Does something and produces output."""

    @abc.abstractmethod
    def evaluate(self, result: Any) -> Result:
        """Evaluate and return a result."""

    def __call__(self, resources: dict) -> Result:
        """Should behave like a standard runnable."""

        self.resources = resources
        self.details = dict()

        intermediate = self.execute()
        if (connect := getattr(self, "connect", None)) is not None:
            intermediate = connect(intermediate)
        result = self.evaluate(intermediate)
        result.details.update(self.details)
        return result



from typing import Union, List
from decimal import Decimal

from ..grader.task import Result, Error, Message


class CleanupResult(Result):
    """Deletion of files, deallocation, etc."""

    def __init__(
            self,
            passing: bool,
            complete: bool = True,
            score: Union[Decimal, int, float, str] = None,
            error: Error = None,
            messages: List[Message] = None,
            details: dict = None):
        """Add fields for expected and actual."""

        super().__init__(
            passing=passing,
            complete=complete,
            score=score,
            error=error,
            messages=messages,
            details=details)




from .grader import Grader
from .grader.task import Error
from .resource import *

from .setup import SetupResult
from .test.correctness import CorrectnessResult
from .test.complexity import ComplexityResult
from .test.memory import MemoryResult
from .cleanup import CleanupResult



from curricula.library.configurable import Configurable
from curricula.library.valgrind import ValgrindReport, run
from ...common.process import ProcessExecutor
from ...common import Evaluator, CompositeRunnable
from . import MemoryResult

import abc


class ValgrindProcessExecutor(ProcessExecutor):
    """Runs valgrind on the executable with given parameters."""

    def execute(self) -> ValgrindReport:
        executable = self.resources[self.executable_name]
        report = run(
            *executable.args,
            *self.args,
            stdin=self.stdin,
            timeout=self.timeout,
            cwd=self.cwd)
        self.details["report"] = report.dump()
        return report


class ValgrindEvaluator(Evaluator, Configurable):
    """Wrap the report as a result.s"""

    def evaluate(self, report: ValgrindReport) -> MemoryResult:
        return MemoryResult.from_valgrind_report(report)


class MemoryRunnable(CompositeRunnable, metaclass=abc.ABCMeta):
    """Override annotation."""

    result_type = MemoryResult


class MemoryTest(ValgrindProcessExecutor, ValgrindEvaluator, MemoryRunnable):
    """Standard Valgrind memory test."""



from typing import Set, Iterable, Iterator

from .collection import TaskCollection
from .dependency import flatten_dependencies
from ...resource import Context

__all__ = ("filter_tasks",)


def filter_problem_specific(prefix: str, collection: Iterable[str]) -> Set[str]:
    """Filter in items prefaced by prefix:xyz as xyz."""

    result = set()
    for item in collection:
        if ":" in item:
            if item.startswith(f"{prefix}:"):
                result.add(item.split(":", maxsplit=1)[1])
        else:
            result.add(item)
    return result


def filter_tasks_by_tag(tags: Set[str], problem_short: str, tasks: TaskCollection) -> Iterator[str]:
    """Return a set of task names by tag."""

    tags = filter_problem_specific(problem_short, tags)
    for task in tasks:
        if not tags.isdisjoint(task.tags):
            yield task.name


def filter_tasks_by_name(names: Set[str], problem_short: str, tasks: TaskCollection) -> Iterator[str]:
    """Filter by name match."""

    names = filter_problem_specific(problem_short, names)

    # Allow wildcard to group by problem
    if "*" in names:
        for task in tasks:
            yield task.name

    # Normal filter
    else:
        for task in tasks:
            if task.name in names:
                yield task.name


def filter_tasks(context: Context, problem_short: str, tasks: TaskCollection) -> Set[str]:
    """Check if a task should be run."""

    if context.task_tags is None and context.task_names is None:
        return set(task.name for task in tasks)

    task_names = set()
    if context.task_tags is not None:
        task_names |= set(filter_tasks_by_tag(context.task_tags, problem_short, tasks))
    if context.task_names is not None:
        task_names |= set(filter_tasks_by_name(context.task_names, problem_short, tasks))

    # Add dependencies
    task_lookup = {task.name: task for task in tasks}
    for task_name in task_names.copy():
        for dependency_task_name in flatten_dependencies(task_name, task_lookup):
            task_names.add(dependency_task_name)

    return task_names



from typing import List, Iterator

from . import Task
from .dependency import topological_sort
from ...exception import GraderException


class TaskCollection:
    """Task collection with convenience tools."""

    tasks: List[Task]

    def push(self, task: Task):
        """Sort dependencies."""

        for existing_task in self.tasks:
            if existing_task.name == task.name:
                raise GraderException(f"duplicate task name \"{task.name}\"")

        self.tasks.append(task)
        topological_sort(self.tasks)

    def __init__(self):
        """Set tasks."""

        self.tasks = []

    def __iter__(self) -> Iterator[Task]:
        """Yield out of list."""

        yield from self.tasks
